from typing import Tuple, Union

import astropy
import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.units import Quantity
from astropy.wcs.utils import WCS, fit_wcs_from_points
from skimage.measure import label, regionprops

from twirl.geometry import pad
from twirl.match import cross_match, find_transform, get_transform_matrix


def _project_tangent_plane(center, skycoords):
    # Calculate the tangent plane projection
    tangent_plane_coords = skycoords.transform_to(center.skyoffset_frame())

    # Access the projected coordinates
    tangent_ra = tangent_plane_coords.lon
    tangent_dec = tangent_plane_coords.lat

    return np.array([tangent_ra.deg, tangent_dec.deg])


def compute_wcs(
    pixel_coords: np.ndarray,
    radecs: np.ndarray,
    tolerance: int = 5,
    quads_tolerance: float = 0.1,
    asterism=4,
    min_match=0.8,
) -> WCS:
    """
    Compute the WCS solution for an image given pixel coordinates and some unordered RA-DEC values.

    Parameters
    ----------
    pixel_coords : np.ndarray
        Pixel coordinates of the sources in the image, shape (n, 2)
    radecs : np.ndarray
        RA-DEC coordinates of the sources in the image, shape (m, 2)
    tolerance : int, optional
        Tolerance for the matching algorithm, by default 5
    asterism : int, optional
        Number of sources to use for matching, by default 4
    min_match : int, optional
        Minimum number of matches required, by default None

    Returns
    -------
    astropy.wcs.WCS
        WCS solution for the image if a match can be computed, None otherwise.
        A match is considered to be computed if at least one source and one target
        star are located less than `tolerance` pixels away from each other.
    """
    original_radecs = radecs.copy()
    center = SkyCoord(*radecs.mean(0), unit="deg")
    radecs = _project_tangent_plane(center, SkyCoord(radecs, unit="deg")).T

    M = find_transform(
        radecs,
        pixel_coords,
        tolerance=tolerance,
        asterism=asterism,
        min_match=min_match,
        quads_tolerance=quads_tolerance,
    )
    if M is None:
        return None
    else:
        radecs_xy = (M @ pad(radecs).T)[0:2].T
        i, j = cross_match(pixel_coords, radecs_xy).T
        M = get_transform_matrix(radecs[j], pixel_coords[i])
        radecs_xy = (M @ pad(radecs).T)[0:2].T
        i, j = cross_match(pixel_coords, radecs_xy).T
        return fit_wcs_from_points(
            pixel_coords[i].T, SkyCoord(original_radecs[j], unit="deg")
        )


def gaia_radecs(
    center: Union[Tuple[float, float], SkyCoord],
    fov: Union[float, Quantity],
    limit: int = 10000,
    circular: bool = True,
) -> np.ndarray:
    """
    Query the Gaia archive to retrieve the RA-DEC coordinates of stars within a given field-of-view (FOV) centered on a given sky position.

    Parameters
    ----------
    center : tuple or astropy.coordinates.SkyCoord
        The sky coordinates of the center of the FOV. If a tuple is given, it should contain the RA and DEC in degrees.
    fov : float or astropy.units.Quantity
        The field-of-view of the FOV in degrees. If a float is given, it is assumed to be in degrees.
    limit : int, optional
        The maximum number of sources to retrieve from the Gaia archive. By default, it is set to 10000.
    circular : bool, optional
        Whether to perform a circular or a rectangular query. By default, it is set to True.

    Returns
    -------
    np.ndarray
        An array of shape (n, 2) containing the RA-DEC coordinates of the retrieved sources in degrees.

    Raises
    ------
    ImportError
        If the astroquery package is not installed.

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> from twirl import gaia_radecs
    >>> center = SkyCoord(ra=10.68458, dec=41.26917, unit='deg')
    >>> fov = 0.1
    >>> radecs = gaia_radecs(center, fov)
    """
    from astroquery.gaia import Gaia

    if isinstance(center, SkyCoord):
        ra = center.ra.deg
        dec = center.dec.deg
    else:
        ra, dec = center

    if not isinstance(fov, u.Quantity):
        fov = fov * u.deg

    if fov.ndim == 1:
        ra_fov, dec_fov = fov.to(u.deg).value
    else:
        ra_fov = dec_fov = fov.to(u.deg).value

    radius = np.min([ra_fov, dec_fov]) / 2

    fields = "ra, dec"

    if circular:
        job = Gaia.launch_job(
            f"select top {limit} {fields} from gaiadr2.gaia_source where "
            "1=CONTAINS("
            f"POINT('ICRS', {ra}, {dec}), "
            f"CIRCLE('ICRS',ra, dec, {radius}))"
            "order by phot_g_mean_mag"
        )
    else:
        job = Gaia.launch_job(
            f"select top {limit} {fields} from gaiadr2.gaia_source where "
            f"ra BETWEEN {ra-ra_fov/2} AND {ra+ra_fov/2} AND "
            f"dec BETWEEN {dec-dec_fov/2} AND {dec+dec_fov/2} "
            "order by phot_g_mean_mag"
        )

    table = job.get_results()
    return np.array([table["ra"].value.data, table["dec"].value.data]).T


def find_peaks(data: np.ndarray, threshold: float = 2.0) -> np.ndarray:
    """
    Find the coordinates of the peaks in a 2D array.

    Parameters
    ----------
    data : np.ndarray
        The 2D array to search for peaks.
    threshold : float, optional
        The threshold (in unit of image standard deviation) above which a pixel is considered
        part of a peak, i.e.
        The default is 2.0.

    Returns
    -------
    np.ndarray
        An array of shape (N, 2) containing the (x, y) coordinates of the N peaks
        found in the input array. The peaks are sorted by decreasing flux.
    """
    threshold = threshold * np.nanstd(data) + np.nanmedian(data)
    regions = regionprops(label(data > threshold), data)
    coordinates = np.array([region.weighted_centroid[::-1] for region in regions])
    fluxes = np.array([np.sum(region.intensity_image) for region in regions])

    return coordinates[np.argsort(fluxes)[::-1]]



from typing import Optional

import numpy as np
from scipy.spatial import cKDTree

from twirl.geometry import get_transform_matrix, pad
from twirl.quads import hashes as hash4
from twirl.triangles import hashes as hash3


def count_cross_match(coords1, coords2, tol=1e-3):
    """
    Counts the number of cross-matches between two sets of 2D points.

    Parameters
    ----------
    coords1 : np.ndarray
        First set of points coordinates, shape (n, 2).
    coords2 : np.ndarray
        Second set of points coordinates, shape (m, 2).
    tol : float, optional
        Tolerance of the match, by default 1e-3.

    Returns
    -------
    int
        Number of cross-matches between coords1 and coords2.
    """
    return np.count_nonzero(
        np.min(np.linalg.norm(coords1[:, None, :] - coords2[None, :, :], axis=-1), 0)
        < tol
    )


def cross_match(coords1, coords2, tolerance=10):
    """
    Finds the closest matches between two sets of 2D points.

    Parameters
    ----------
    coords1 : np.ndarray
        First set of points coordinates, shape (n, 2).
    coords2 : np.ndarray
        Second set of points coordinates, shape (m, 2).
    tolerance : float, optional
        Tolerance of the match, given in coords1 points units, by default 10.

    Returns
    -------
    np.ndarray
        Array of matched indices, where each row contains the indices of the matched points in coords1 and coords2.
    """

    matches = []

    for i, s in enumerate(coords1):
        distances = np.linalg.norm(s - coords2, axis=1)
        closest = np.argmin(distances)
        if distances[closest] < tolerance:
            matches.append([i, closest])

    return np.array(matches)


def find_transform(
    radecs: np.ndarray,
    pixels: np.ndarray,
    min_match: float = 0.7,
    asterism: int = 4,
    quads_tolerance: float = 0.02,
    tolerance: float = 12,
) -> np.ndarray:
    """
    Finds the transformation matrix from `radecs` to `pixels`.

    Parameters
    ----------
    radecs : np.ndarray
        The coordinates to be transformed, shape (n, 2).
    pixels : np.ndarray
        The target coordinates, shape (m, 2).
    min_match : float, optional
        The fraction of `pixels` coordinates that must be matched to stop the search.
        I.e., if the number of matched points is `>= min_match * len(pixels)`, the
        search stops and return the found transform. By default 0.7.
    asterism : int, optional
        The asterism to use for hashing, either 3 or 4, by default 4.
    quads_tolerance : float, optional
        The minimum euclidean distance between two quads to be matched and tested.
        By default 0.02.
    tolerance : float, optional
        The minimum distance between two coordinates to be considered cross-matched
        (in `pixels` units). This serves to compute the number of coordinates being
        matched between `radecs` and `pixels` for a given transform.
        By default 12.

    Returns
    -------
    np.ndarray
        The transformation matrix from `radecs` to `pixels`.
    """

    if asterism == 3:
        asterism_function = hash3
    elif asterism == 4:
        asterism_function = hash4
    else:
        raise ValueError("available asterisms are 3 and 4")

    hashes_pixels, asterism_pixels = asterism_function(pixels)
    hashes_radecs, asterism_radecs = asterism_function(radecs)

    tree_pixels = cKDTree(hashes_pixels)
    tree_radecs = cKDTree(hashes_radecs)
    pairs = []

    ball_query = tree_pixels.query_ball_tree(tree_radecs, r=quads_tolerance)
    matches = []

    for i, j in enumerate(ball_query):
        if len(j) > 0:
            pairs += [[i, k] for k in j]

    for i, j in pairs:
        M = get_transform_matrix(asterism_radecs[j], asterism_pixels[i])
        test = (M @ pad(radecs).T)[0:2].T
        match = count_cross_match(pixels, test, tolerance)
        matches.append(match)

        if min_match is not None:
            if match >= min_match * len(pixels):
                break

    if len(matches) == 0:
        return None
    else:
        i, j = pairs[np.argmax(matches)]
        M = get_transform_matrix(asterism_radecs[j], asterism_pixels[i])
        return M



from itertools import combinations

import numpy as np

from twirl.geometry import proj, u1u2


def reorder(quads):
    distances = np.linalg.norm(
        quads[:, :, :, None] - np.rollaxis(quads.T, 2)[:, None, :, :], axis=2
    )
    i = np.argmax(np.max(distances, 1), 1)
    idxs = np.roll(
        np.argsort(distances[range(len(quads)), i], axis=1)[:, ::-1], 1, axis=1
    )
    return np.array([q[i] for q, i in zip(quads, idxs)])


def good_quads(quads, circletol=0.01):
    """
    whether all points are contained in a circle (see Lang2009)
    """
    q0, q1 = quads[:, 0], quads[:, 1]
    r = np.linalg.norm(q1 - q0, axis=1) / 2
    center = q0 + (q1 - q0) / 2
    in_circle = np.linalg.norm(quads - center[:, None, :], axis=2) <= r[:, None] * (
        1 + circletol
    )
    return np.all(in_circle, axis=1)


def quad_hash(quads):
    a, b, c, d = np.rollaxis(quads, 1)

    norm = np.linalg.norm(b - a, axis=1)

    def u1u2(a, b):
        """compute x, y basis as defined in Lang2009"""
        co_m = np.cos(-np.pi / 4)
        si_m = np.sin(-np.pi / 4)
        co_p = np.cos(np.pi / 4)
        si_p = np.sin(np.pi / 4)
        rm = np.array([[co_m, -si_m], [si_m, co_m]])
        rp = np.array([[co_p, -si_p], [si_p, co_p]])
        x = (rm @ (b - a).T).T + a
        y = (rp @ (b - a).T).T + a
        return x, y

    u1, u2 = u1u2(a, b)

    def proj(p, origin, axe):
        """projection of a point p on a segment from origin to axe"""
        n = axe - origin
        n /= np.linalg.norm(n, axis=1)[:, None]
        return np.sum((p - origin) * n, 1)

    return (
        np.array(
            [
                proj(c, a, u1) / norm,
                proj(d, a, u1) / norm,
                proj(c, a, u2) / norm,
                proj(d, a, u2) / norm,
            ]
        ).T,
        np.rollaxis(np.array([a, b]), 1),
    )


def clean_quads(xy):
    assert xy.shape[1] == 2
    quads_idxs = np.array(list(combinations(np.arange(xy.shape[0]), 4)))
    quads = xy[quads_idxs]
    ordered_quads = reorder(quads)
    return ordered_quads[good_quads(ordered_quads)]


def hashes(xy):
    assert xy.shape[1] == 2
    quads_idxs = np.array(list(combinations(np.arange(xy.shape[0]), 4)))
    quads = xy[quads_idxs]
    ordered_quads = reorder(quads)
    good_ordered_quads = ordered_quads[good_quads(ordered_quads)]
    h, u = quad_hash(good_ordered_quads)
    # we sort hashes from larger AB (see Lang 2008)
    idxs = np.argsort(np.linalg.norm(u[:, 1] - u[:, 0], axis=1))
    return h[idxs[::-1]], good_ordered_quads[idxs[::-1]]



from itertools import combinations

import numpy as np

from twirl.geometry import triangle_angles


def order_points(triangles):
    """
    Orders the vertices of each triangle in a consistent manner.

    Vertices are ordered by their distance to the centroid of the triangle.

    Parameters
    ----------
    triangles : ndarray
        An array of shape (n_triangles, 3, 2) representing the vertices of each triangle.

    Returns
    -------
    ndarray
        An array of shape (n_triangles, 3, 2) with the vertices of each triangle ordered consistently.
    """

    # Compute the centroid of the triangles
    centroids = np.mean(triangles, axis=1, keepdims=True)

    # Compute the distances from the centroid to the vertices
    distances = np.linalg.norm(triangles - centroids, axis=-1)

    # Get the indices that would sort the distances
    sort_indices = np.argsort(distances, axis=1)

    # Use numpy's advanced indexing to sort the triangles
    ordered_triangles = np.take_along_axis(triangles, sort_indices[:, :, None], axis=1)

    return ordered_triangles


def hashes(xy, min_angle=np.deg2rad(30)):
    """
    Computes the hashes of the triangles formed by the points in xy.

    Parameters
    ----------
    xy : ndarray
        An array of shape (n_points, 2) representing the x and y coordinates of each point.
    min_angle : float, optional
        The minimum angle (in radians) that a triangle must have to be included in the hashes. Default is 30 degrees.

    Returns
    -------
    hashes : ndarray
        An array of shape (n_hashes, 2) representing the hashes of the triangles.
    triangles : ndarray
        An array of shape (n_triangles, 3, 2) representing the vertices of each triangle.
    """

    triangles_idxs = np.array(list(combinations(np.arange(xy.shape[0]), 3)))
    triangles = xy[triangles_idxs]
    triangles = order_points(triangles)
    angles = triangle_angles(triangles)
    # keep only triangles with any angle > min_angle
    mask = np.all(np.abs(angles) > min_angle, axis=1)
    triangles = triangles[mask]
    angles = angles[mask]
    hashes = np.sort(angles, axis=1)[:, 0:2]
    return hashes, triangles
